#! /usr/bin/python3

import sys, os, re
# allow local modules like ressource_rc
sys.path.append(os.path.dirname(__file__))

from PyQt5.QtWidgets import QApplication, QMainWindow, QDialog, \
    QAbstractScrollArea, QPushButton
from PyQt5.QtCore import QTranslator, QLocale, Qt, pyqtSignal, QLibraryInfo

from ui_live_clone import Ui_MainWindow
from ui_about import Ui_Dialog as Ui_AboutDialog
from ui_packageEditor import Ui_PackageEditor

from dbus.mainloop.glib import DBusGMainLoop, threads_init
import gi
gi.require_version('UDisks', '2.0')
import usbDisk2

from tools import Tool, MountPoint, FileTreeSelectorModel, MyItem
from wizards import ToolWizard, CloneWizard, RunWizard
from monitor import Monitor
from subprocess import Popen, PIPE
from markdown import markdownFromFile

from datetime import datetime
import io
from subprocess import call

# a pattern for the output of dd when copying the ISO image
# this pattern reports group(1) = bytes copied so far,
# group(2) = duration in seconds, group(3) = copy speed in GB/s
ddPattern=re.compile(r"^(\d+).*\(.*\).*, (\d+) s, ([.,\d]+) .*/s *$")

class MyMain(QMainWindow):
    """
    La fenêtre principale de l'application
    """

    keysChanged        = pyqtSignal(str) # means that some USB stick has changed
    newmonitor        = pyqtSignal(Monitor)
    monitorfinished   = pyqtSignal(Monitor)
    monitorclosed     = pyqtSignal(Monitor)
    monitorout        = pyqtSignal(Monitor, str)
    monitorerr        = pyqtSignal(Monitor, str)
    
    def __init__(self, lesCles, owndisk, parent=None):
        """
        Constructeur de la fenêtre principale
        :param lesCles: une instance de usbDisk2.Available
        :param owndisk: une chaîne vide ou quelque chose comme "sdc"
        :type  owndisk: str
        :param parent: fenêtre parente, None par défaut
        """
        QMainWindow.__init__(self, parent)
        self.lesCles=lesCles
        self.owndisk=owndisk
        self.wd=os.path.abspath(os.path.dirname(__file__))
        self.monitors={}
        self.ui=Ui_MainWindow()
        self.ui.setupUi(self)
        self.initTable()
        self.initTab()
        self.initHelpTab()
        
        self.about=QDialog()
        self.about.ui=Ui_AboutDialog()
        self.about.ui.setupUi(self.about)
        self.initAbout()

        self.ui.helpButton.clicked.connect(self.showHelp)
        self.ui.quitButton.clicked.connect(self.close)
        self.ui.toolsButton.clicked.connect(self.toolAction)
        self.ui.cloneButton.clicked.connect(self.cloneAction)
        self.ui.cloneButton1.clicked.connect(self.cloneAction)
        self.ui.runButton.clicked.connect(self.runAction)
        
        self.keysChanged.connect(self.update_keys)
        self.newmonitor.connect(self.new_monitor)
        self.monitorfinished.connect(self.monitor_finished)
        self.monitorclosed.connect(self.monitor_closed)
        self.monitorout.connect(self.monitor_out)
        self.monitorerr.connect(self.monitor_err)
        self.update_keys() # check already connected USB sticks
        return

    def runAction(self):
        """
        Lance un virtualiseur qemu-kvm avec un des disques
        """
        devices=self.not_self_keys(format="udev")
        wiz=RunWizard(self, devices, self.lesCles)
        is_ok = wiz.exec_()
        disks=wiz.field("disks")
        result={}
        if is_ok:
            for d in disks.split(","):
                result[d]=Tool().runDisk(d, self)
        return result

    def cloneAction(self, event):
        """
        Fonction de rappel pour self.ui.toolsButtoncloneButton ; sert à cloner
        les disques
        """
        wiz=CloneWizard(self)
        wiz.exec_()
        return
        
        
    def toolAction(self, event):
        """
        Fonction de rappel pour self.ui.toolsButton
        """
        wiz=ToolWizard(self)
        wiz.exec_()
        return

    def showHelp(self):
        """
        Montre l'aide intégrée
        """
        self.about.show()
        self.ui.tabWidget.setCurrentIndex(0) # ensure the help tab is visible
        return

    def rindexFromDev(self, dev):
        """
        :param dev: un périphérique comme "/dev/sdc"
        :type  dev: str
        :return: l'index de la ligne du widget table qui contient ce
        périphérique, ou None s'il n'est pas là.
        """
        t=self.ui.tableWidget
        rindex=0
        found=False
        for rindex in range(t.rowCount()):
            if t.item(rindex,0).data(0) == dev:
                found=True
                break
        if not found:
            return None
        return rindex
    
    def mdToEdit(self, path, edit):
        """
Remplit une instance de QTextEdit avec un fichier Markdown
        :param path: le chemin du fichier .md, relatif à ce fichier source.
        :type  path: str
        :param edit: le widget d'affichage
        :type  edit: QTextEdit
        """
        path=os.path.join(self.wd, path)
        html=io.BytesIO()
        markdownFromFile(input=path, output=html)
        html.seek(0)
        edit.setHtml(html.read().decode("utf-8"))
        return        

    def initAbout(self):
        """
        Remplit le dialogue "À propos" avec un texte localisé
        """
        self.mdToEdit(self.tr("HelpAbout.md"), self.about.ui.textEdit)
        return
    
    def initHelpTab(self):
        """
        Met en place le fichier d'aide dans le bon onglet
        """
        self.mdToEdit(self.tr("HelpTab.md"), self.ui.textEdit)
        return
    
    def monitor_finished(self, monitor):
        """
        Fonction de rappel pour quand un monitor a fini son job.
        """
        t=self.ui.tableWidget
        t.setItem(monitor.rindex,2,
                  MyItem(self.tr("Clone ready ... Tab #{0}").format(monitor.index)))
        t.resizeColumnsToContents()
        return

    def new_monitor(self, monitor):
        """
        Fonction de rappel pour quand un nouveau moniteur devient prêt
        :param monitor: le nouveau moniteur
        :type  monitor: Monitor
        """
        self.update_monitored_row(monitor)
        return
        
    def monitor_closed(self, monitor):
        """
        Fonction de rappel quand un moniteur est refermé
        :param monitor: le nouveau moniteur
        :type  monitor: Monitor
        """
        device = self.monitors.pop(monitor)
        w=self.ui.tabWidget.widget(monitor.index)
        self.ui.tabWidget.removeTab(monitor.index)
        if w: w.close()
        t=self.ui.tableWidget
        t.setItem(monitor.rindex,2,
                  MyItem(self.tr("Tab closed.")))
        t.resizeColumnsToContents()
        return

    def monitor_out(self, monitor, text):
        """
        Callback function when a monitor sends a text which appeared in stdout
        """
        print("out", monitor, text)
        return
    
    def monitor_err(self, monitor, text):
        """
        Callback function when a monitor sends a text which appeared in stderr
        """
        if monitor.category == "clone_monitor":
            msg = monitor.progresstext(text)
            if msg:
                self.update_monitored_row(monitor, msg)
        return
    

    def makeRow(self, rindex, disk, already):
        """
        Crée une ligne dans la vue table pour un disque
        :param index: numéro de la ligne
        :param disk: l'instance de disque
        :param already: est Vrai si on sait que la première colonne est déjà
        correcte, et que la ligne existe déjà.
        :param dic: un dictionnaire disque => partitions
        :return: une nouvelle valeur pour l'index, augmentée si une ligne a été
        créée.
        """

        def description(partition_ud):
            dev=os.path.basename(partition_ud.path)
            label=partition_ud.label
            label1=label.split(" ")[0]
            if len(label) > 11:
                label=label1[:8]+"..."
            if partition_ud.label:
                return "{}({})".format(dev, label)
            else:
                return dev
            
        t=self.ui.tableWidget
        dic=self.lesCles.disksDict()
        nextIndex=rindex
        shortDisk=os.path.join("/dev",os.path.basename(disk))
        shortParts=self.lesCles.parts_summary(disk)
        if not already:
            t.insertRow(rindex)
            t.setItem(rindex,0,MyItem(shortDisk))
            nextIndex=rindex+1
            # create the button if it is not already there
            button=QPushButton(self.tr("Clone to {}").format(shortDisk))
            b_function=lambda:self.clonage("/dev/{}".format(shortDisk), button, rindex)
            button.clicked.connect(b_function)
            t.setCellWidget(rindex,3,button)
        t.setItem(rindex,1,MyItem(", ".join(shortParts)))
        ## check whether there is still a Monitor with a process for this device
        for m, devicepath in self.monitors.items():
            if shortDisk == os.path.basename(devicepath):
                self.update_monitored_row(m, button)
        return nextIndex

    def update_monitored_row(self, m, msg = ""):
        """
        Quelques trucs à lancer pour la mise à jour d'une ligne de tableau
        si elle est déjà en relation à un quelconque moniteur
        :param m: le nouveau moniteur
        :type  m: Monitor
        :param msg: an extra message (defaults to "")
        """
        t=self.ui.tableWidget
        r=m.rindex
        i=m.index
        t.setItem(
            r, 2,
            MyItem(self.tr("{} ... Tab #{} {}").format(m.actionMessage, i, msg))
        )
        t.resizeColumnsToContents()
        return

    def not_self_keys(self, format=None):
        """
        :param format: si ça vaut "udev" on renverra un chemin udev plutôt
        qu'un chemin uDisks2
        :return: une liste de chemins de disques USB, sauf pour le disque
        USB qui porte éventuellement le système actuellement en fonction.
        """
        if format=="udev":
            return [os.path.join("/dev", os.path.basename(d)) \
                    for d in  sorted(self.lesCles.disksDict()) \
                    if not self.owndisk or self.owndisk not in d]
        # return the uDisks path by default
        return [d for d in  sorted(self.lesCles.disksDict()) \
                if not self.owndisk or self.owndisk not in d]
    
    def update_keys(self):
        """
        Fonction de rappel pour quand un disque ou une clé USB est
        connectée ou déconnectée ou quand ses partitions sont modifiées.
        """
        # builds the list of USB sticks, except the stick enventually
        # used to boot the system
        disks=self.not_self_keys()
        t=self.ui.tableWidget
        rindex=0
        ## first, delete rows with disk references which are no longer valid
        for rindex in list(range(t.rowCount()))[::-1]:
            if t.item(rindex,0).data(0) not in disks:
                t.removeRow(rindex)
        ## then, add new disks and update existing ones
        if disks:
            self.statusBar().showMessage(self.tr("Current USB disks: {}").format(", ".join([os.path.basename(d) for d in disks])))
        else:
            self.statusBar().showMessage(self.tr("No USB stick. Please plug a USB flash disk."))
        for disk in disks:
            devDisk=os.path.join("/dev", os.path.basename(disk))
            while rindex < t.rowCount() and \
                  t.item(rindex,0).data(0) < devDisk:
                ## skip the lines before the place of devDisk
                rindex+=1
            already = rindex < t.rowCount() and \
                t.item(rindex,0).data(0) == devDisk
            rindex=self.makeRow(rindex, disk, already)
        t.resizeColumnsToContents()
        return

    def initTable(self):
        """
        Initialise le tableau des périphériques
        (autres que la clé USB vive éventuellement bootée)
        """
        t=self.ui.tableWidget
        t.setColumnCount(3)
        t.setHorizontalHeaderLabels([
            self.tr("Device"), self.tr("Partitions"),self.tr("Status")
        ])
        t.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        return

    def initTab(self):
        """
        Initialise le QTabWidget dans l'interface utilisateur
        retire le widget numéro 1, qui provient du Designer.
        Attention, le widget numéro 0 est spécial : il contient l'aide.
        """
        w=self.ui.tabWidget.widget(1)
        self.ui.tabWidget.removeTab(1)
        w.close()
        return

def main():
    """
    Entrée principale du programme
    """
    app = QApplication(sys.argv)
    # i18n stuff
    locale = QLocale.system().name()
    translation="live_clone_{}.ts".format(locale)
    langPath=os.path.join(os.path.abspath(os.path.dirname(__file__)),"lang",translation)
    translator = QTranslator(app)
    translator.load(langPath)
    app.installTranslator(translator)
    t1=QTranslator(app)
    t1.load(QLocale(), "qt", "_", QLibraryInfo.location(QLibraryInfo.TranslationsPath))
    app.installTranslator(t1)
    t2=QTranslator(app)
    t2.load(QLocale(), "qtbase", "_", QLibraryInfo.location(QLibraryInfo.TranslationsPath))
    app.installTranslator(t2)
    #############
    lesCles=usbDisk2.Available()
    p=Popen("ls /usr/lib/live/mount/persistence", shell=True, stdout=PIPE, stderr=PIPE)
    out, err = p.communicate()
    shortDiskNames=out.decode("utf8").split()
    ownDisk=""
    if shortDiskNames:
        # if we have ['sda1', 'sda3'], this should give "sda"
        ownDisk=shortDiskNames[0][:-1]
    w = MyMain(lesCles, ownDisk)
    # addHook is not designed to work with a method inside an object
    # so, let us define the hook outside the main window, and relay
    # a signal to the main window
    # the slot keysChanged in MyMain accepts a string parameter
    # which is the current device if the system is booted from it
    def show_keys(man, obj):
        if lesCles.modified: # filter to detect only USB stick objects
            w.keysChanged.emit(ownDisk)
        lesCles.modified=False
        return
    lesCles.addHook('object-added',   show_keys)
    lesCles.addHook('object-removed', show_keys)
    
    w.show()
    
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



licence={}
licence_en="""
    file usbDisk2.py
    This file was part of the project scolasync.
    It was rewritten to update some parts and make it useful for
    the project Live-Clone
    
    Copyright (C) 2014-2019 Georges Khaznadar <georgesk@debian.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

licence['en']=licence_en
dependences="python3-dbus python3-dbus.mainloop.qt"


import dbus, subprocess, os, os.path, re, time, threading, logging, inspect
from dbus.mainloop.glib import DBusGMainLoop, threads_init
from gi.repository import Gio, GLib, UDisks
from PyQt5.QtWidgets import *

#################### activate debugging #######################
debug=False # or not

def inspectData0():
    """
    Here is a version of inspectData which does nothing
    :return: an empty string
    """
    return ""

def inspectData1():
    """
    Here is a version of inspectData which does the job
    :return: a string with useful debugging information
    """
    caller=1
    callerframerecord = inspect.stack()[caller]
    frame = callerframerecord[0]
    info = inspect.getframeinfo(frame)
    return " -- file={0}, function={1}, line={2}".format(
        info.filename, info.function, info.lineno
        )

inspectData = inspectData0 # dummy function by default
if debug :
    logging.basicConfig(level=logging.DEBUG)
    inspectData = inspectData1 # effective function
else:
    pass
    # logging.basicConfig(level=logging.NOTSET)
###############################################################


def safePath(obj):
    """
    Récupère de façon sûre le path d'une instance de UDisksObjectProxy

    :param obj: quelque chose comme 
    :type  obj: instance de UDisksObjectProxy, ou simple chaine
    :return: le path d'une instance de UDisksObjectProxy
    :rtype: str
    """
    if type(obj)==type(""):
        path=obj
    else:
        path= obj.get_object_path()
    posUnderscore=path.rfind("_")
    posSlash=path.rfind("/")
    if posUnderscore > posSlash: # il faut retirer tout après l'underscore final
        path=path[:posUnderscore]
    return path
                
def fs_size(device):
    """
    Renvoie la taille d'un système de fichier et la place disponible
    :param device: la désignation d'un périphérique, "/dev/sdc" par exemple
    :type  device: str
    :return: un tuple : taille totale et espace libre
    :rtype: tuple(int, int)
    """
    try:
        stat = os.statvfs(device)
    except:
        return (0, 0)
    free = stat.f_bsize * stat.f_bavail # les blocs réservés sont inclus
    total = stat.f_bsize * stat.f_blocks
    return (total, free)


############ la variable suivante a été recopiées à l'aveugle ################
############ depuis un fichier du projet USBcreator ##########################
no_options = GLib.Variant('a{sv}', {})
##############################################################################
##############################################################################

######### des "chemins" correspondant à des disques non débranchables ########
not_interesting = (
    # boucle
    '/org/freedesktop/UDisks2/block_devices/loop',
    # disque raid
    '/org/freedesktop/UDisks2/block_devices/dm_',
    # mémoire vive
    '/org/freedesktop/UDisks2/block_devices/ram',    
    '/org/freedesktop/UDisks2/block_devices/zram',
    # disques durs
    '/org/freedesktop/UDisks2/drives/', 
    )

class UDisksBackend:
    """
    Cette classe a été inspirée par le projet USBcreator. 
    Plusieurs modifications ont été faites au code original.
    Les fonctions de rappel ne tiennent compte que des périphériques USB,
    et pas des disques branchés sur d'autres bus.
    """
    
    def __init__(self, logger=logging, diskClass=object):
        """
        Le constructeur.
        :param logger: un objet permettant de journaliser les messages ; 
        par défaut il se confond avec le module logging
        :type logger: un module tel que logging
        :param diskClass: la classe à utiliser pour créer des instances de disques
        :type diskClass: class
        """
        self.install_thread = None
        self.logger=logger
        ## self.targets est un dictionnaire des disques détectés
        ## les clés sont les paths et les contenus des instances de diskClass
        self.diskClass=diskClass
        self.targets = {}
        ## self.modified signifie une modification récente, à prendre en compte
        ## par une application au niveau utilisateur
        self.modified=False
        DBusGMainLoop(set_as_default=True)
        threads_init()
        self.bus = dbus.SystemBus()
        self.udisks = UDisks.Client.new_sync(None)
        self.manager = self.udisks.get_object_manager()
        self.cbHooks = {
            'object-added':      {
                'profile': ['man', 'obj'],
                'hooks'  : []
                },
            'object-removed':      {
                'profile': ['man', 'obj'],
                'hooks'  : []
                },
            'interface-added':      {
                'profile': ['man', 'obj'],
                'hooks'  : []
                },
            'interface-removed':      {
                'profile': ['man', 'obj'],
                'hooks'  : []
                },
            'interface-proxy-properties-changed':      {
                'profile': ['man', 'obj', 'interface'],
                'hooks'  : []
                },
        }
        # mise en place des fonctions de rappel à utiliser pour tout changement
        self.addHook('object-added',
                     lambda man, obj: self._udisks_obj_added(obj))
        self.addHook('object-removed',
                     lambda man, obj: self._udisks_obj_removed(obj))
        self.addHook('interface-added',
                     lambda man, obj, iface: self._device_changed(obj))
        self.addHook('interface-removed',
                     lambda man, obj, iface: self._device_changed(obj))
        self.addHook('interface-proxy-properties-changed',
                     lambda man, obj, iface, props, invalid: self._device_changed(obj))

    def addHook(self, signal, func):
        """
        Ajoute une fonction à appeler pour un signal nommé, et enregistre
        cette fonction dans self.cbHooks, après vérification de sa liste
        de paramètres.
        :param signal: le nom d'un signal
        :type signal: str
        :param func: une fonction
        :type  func: fonction sans paramètres
        :return: le résultat de l'appel à self.manager.connect(signal,func)
        """
        if inspect.getargspec(func).args == self.cbHooks[signal]['profile']:
            cb=self.manager.connect(signal,func)
            self.cbHooks[signal]['hooks'].append(cb)
            return cb
        return None
    
    # voir le fichier integration-test issu des sources de udisks2
    def retry_mount(self, fs, timeout=5, retryDelay=0.3):
        """
        Essaie de monter un système de fichier jusqu'à ce qu'il
        cesse d'échouer avec "Busy", ou que l'erreur soit "déjà monté".
        Échoue si l'erreur est autre que les deux précédentes.
        :param fs: un système de fichier à monter ; par exemple, "/dev/sdc1".
        :param timeout: nombre de secondes d'attente au maximum
        :type  timeout: float
        :param retryDelay: délai entre deux essais (en secondes)
        :type  retryDelay: float
        :return: le résultat de fs.call_mount_sync(no_options, None) en cas de réussite, sinon une chaîne vide
        """
        while timeout >= 0:
            try:
                return fs.call_mount_sync(no_options, None)
            except GLib.GError as e:
                if 'UDisks2.Error.AlreadyMounted' in e.message:
                    m=re.match(r".*already mounted[^/]*([^\']+).*",e.message)
                    return m.group(1)
                elif 'UDisks2.Error.DeviceBusy' in e.message:
                    pass
                else:
                    raise
                time.sleep(retryDelay)
                timeout -= retryDelay
        return ''

    def detect_devices(self):
        """
        Fait un inventaire des disques. Modifie une propriété de l'instance.
        """
        for obj in self.manager.get_objects():
            self._udisks_obj_added(obj)

    def _interesting_obj(self, obj):
        """
        Décide si un objet est intéressant à cataloguer, donc rejette ce
        qui ne se branche pas sur le bus USB.
        :param obj: l'objet qu'on examine
        :type  obj: une instance de UDisksObjectProxy
        :return: un triplet (interesting (bool), drive (objet),
        partition (objet)).
        """
        interesting=False
        drive=None
        partition=None
        
        # ne tient pas compte des périphériques non débranchables
        path = safePath(obj)
        for boring in not_interesting:
            if path.startswith(boring):
                return interesting, drive, partition

        # ne considère que les périphériques de type block
        block = obj.get_block()
        if not block:
            return interesting, drive, partition
        
        # initialise drive, nom du disque ?
        drive_name = block.get_cached_property('Drive').get_string()
        if drive_name == '/':
            return interesting, drive, partition
        else:
            drive = self.udisks.get_object(drive_name).get_drive()
        
        # on ne tient pas compte des CDROMS ni DVDROMS
        if drive and drive.get_cached_property('Optical').get_boolean():
            return interesting, drive, partition

        interesting=True
        # détermine si on a un disque ou une partition
        partition = obj.get_partition()
        return interesting, drive, partition
    
    def _udisks_obj_added(self, obj):
        """
        Fonction de rappel pour les ajouts de disque
        :param obj: un objet renvoyé par l'évènement
        """
        interesting, drive, part = self._interesting_obj(obj)
        if part:
            self._udisks_partition_added(obj, drive, part)
        elif drive:
            self._udisks_drive_added    (obj, drive, part)
        return

    def objIsUsb(self,obj):
        """
        Décide si un périphérique est de type USB
        :param obj: l'objet qui nous intéresse
        :type  obj: UDisksObjectProxy
        :return: vrai si c'est un périphérique USB
        :rtype: bool
        """
        if obj.get_block().props.read_only:
            return False # on ne considère pas les disques non inscriptibles
        for s in obj.get_block().get_cached_property('Symlinks'):
            if b'/dev/disk/by-id/usb' in bytes(s):
                return True
        return False
    
    def _udisks_partition_added(self, obj, drive, partition):
        """
        Fonction de rappel pour l'ajout d'une partition,
        met à jour self.targets
        :param obj: l'objet qui nous intéresse
        : type  obj: UDisksObjectProxy
        :param drive: une instance de ...
        :param partition: une instance de ...
        """
        path  = safePath(obj)
        block = obj.get_block()
        self.logger.debug(QApplication.translate("uDisk","Added partition %s",None) % path+inspectData())
        fstype = block.get_cached_property('IdType').get_string()
        parent = partition.get_cached_property('Table').get_string()
        total = drive.get_cached_property('Size').get_uint64()
        free = -1
        mount = ''
        fs = obj.get_filesystem()
        if fs:
            mount_points = fs.get_cached_property('MountPoints').get_bytestring_array()
            if len(mount_points)>0:
                mount= mount_points[0]
            if not mount and fstype == 'vfat':
                try:
                    mount = self.retry_mount(fs)
                except:
                    logging.exception(QApplication.translate("uDisk","Failed to mount the disk: %s",None) % path)
        if mount:
            total, free = fs_size(mount)
        isUsb=self.objIsUsb(obj)
        if not isUsb:
            self.logger.debug(QApplication.translate("uDisk","Disk not added: not a USB partition",None)+inspectData())
        elif total < 1:
            self.logger.debug(QApplication.translate("uDisk","Disk not added: empty partition",None)+inspectData())
        else:
            udisk=self.diskClass(
                path=path, mp=mount, isUsb=isUsb,
                vendor=drive.get_cached_property('Vendor').get_string(),
                model=drive.get_cached_property('Model').get_string(),
                parent=safePath(parent),
                fstype=fstype,
                serial=block.get_cached_property('Drive').get_string().split('_')[-1],
                uuid=block.get_cached_property('IdUUID').get_string(),
                free=free,
                capacity=total,
                device=block.get_cached_property('Device').get_bytestring().decode('utf-8'),
                label=block.get_cached_property('IdLabel').get_string()
            )
            self.targets[path] = udisk
            self.modified=True
        return
             
    def _udisks_drive_added(self, obj, drive, part):
        """
        Fonction de rappel pour l'ajout d'un périphérique "disque"
        :param obj: l'objet qui nous intéresse
        : type  obj: UDisksObjectProxy
        :param drive: une instance de ...
        :param part: une instance de ...
        """
        path  = safePath(obj)
        block = obj.get_block()
        if path in self.targets:
            self.logger.debug(QApplication.translate("uDisk","Already added disk: %s",None) % path+inspectData())
            return
        self.logger.debug(QApplication.translate("uDisk","Added disk: %s",None) % path+inspectData())
        size = drive.get_cached_property('Size').get_uint64()
        ##### désactivé, quelquefois drive.get_cached_property('Size').get_uint64()
        ##### renvoie des résultats erronés juste après le branchement
        """
        if size <= 0:
            self.logger.debug(QApplication.translate("uDisk","Disk not added: zero byte partition.",None)+inspectData())
            return
        """
        isUsb = self.objIsUsb(obj)
        if not isUsb:
            self.logger.debug(QApplication.translate("uDisk","Disk not added: not a USB partition",None)+inspectData())
        else:
            udisk=self.diskClass(
                path=path,
                isUsb=isUsb,
                parent='',
                vendor=drive.get_cached_property('Vendor').get_string(),
                model=drive.get_cached_property('Model').get_string(),
                serial=block.get_cached_property('Drive').get_string().split('_')[-1],
                uuid=block.get_cached_property('IdUUID').get_string(),
                capacity=size,
                device=block.get_cached_property('Device').get_bytestring().decode('utf-8'),
            )
            self.targets[path] =udisk
            self.modified=True
        return
        
    def _device_changed(self, obj):
        """
        Fonction de rappel pour prendre en compte une modification
        d'un périphérique
        :param obj: l'objet qui nous intéresse
        """
        path = safePath(obj)
        self.logger.debug(QApplication.translate("uDisk","Change for the disk %s",None) % path+inspectData())

    def _udisks_obj_removed(self, obj):
        """
        Fonction de rappel déclenchée par le retrait d'un disque.
        Met à jour self.targets
        :param obj: l'objet qui nous intéresse
        :type  obj: UDisksObjectProxy
        """
        path=safePath(obj)
        logging.debug(QApplication.translate("uDisk","Disk unplugged from the system: %s",None) % path)
        if path in self.targets:
            self.targets.pop(path)
            self.modified=True

class uDisk2:
    """
    une classe pour représenter un disque ou une partition.

    les attributs publics sont :
    -  path  le chemin dans le système dbus
    -  device l'objet dbus qui correspond à l'instance
    -  device_prop un proxy pour questionner cet objet dbus
    -  selected booléen vrai si on doit considérer cette instance comme sélectionnée. Vrai à l'initialisation
    -  rlock un verrou récursif permettant de réserver l'usage du media pour un seul thread
    """

    def __init__(
            self,
            path,
            mp='',
            isUsb=False,
            vendor='',
            model='',
            parent=None,
            fstype='',
            serial='',
            uuid='',
            free=0,
            capacity=0,
            device='',
            firstFat=None,
            selected=True,
            label=''
    ):
        """
        Le constructeur
        :param path: un chemin comme '/org/freedesktop/UDisks2/block_devices/sdX'
        :type  path: str
        :param mp: point de montage ('' par défaut)
        :type  mp: str
        :param isUsb: en général, vrai vu qu'on se s'intéressera qu'à des périphériques USB
        :type  isUsb: bool
        :param vendor: indication de vendeur
        :type  vendor: str
        :param model: indication de modèle
        :type  model: str
        :param parent: périphérique parent (None par défaut)
        :param fstype: type de système de fichiers
        :type  fstype: str
        :param serial: numéro de série
        :type  serial: str
        :param uuid: identifiant donné au disque lors du formatage
        :type  uuid: str
        :param free: taille de la zone libre pour l'écriture
        :type  free: int
        :param capacity: taille du périphérique
        :type  capacity: int
        :param device: pseudo-fichier pour l'accès au périphérique, par exemple "/dev/sdc1"
        :type  device: str
        :param firstFat: une instance de uDisk2, de type vfat parmi les partitions ; utile seulement pour les disques sans table de partition
        :param selected: vrai/faux selon qu'on sélectionne ou non le périphérique (vrai par défaut)
        :type  selected: bool
        :param label: le label s'il y en a un pour le système de fichiers
        :type  label: str
        """
        self.path=path
        self.mp=mp
        self.isUsb=isUsb
        self.vendor=vendor
        self.model=model
        self.parent=parent
        self.fstype=fstype
        self.stickid=serial
        self.uuid=uuid
        self.free=free
        self.capacity=capacity
        self.devStuff=device
        self.firstFat=firstFat
        self.selected=selected
        self.label=label
        self.rlock=threading.RLock()
        return
            
    _itemNames={
        "1mp":QApplication.translate("uDisk","mount point",None),
        "2capacity":QApplication.translate("uDisk","size",None),
        "3vendor":QApplication.translate("uDisk","brand",None),
        "4model":QApplication.translate("uDisk","disk model",None),
        "5stickid":QApplication.translate("uDisk","serial number",None),
        }

    _specialItems={"0Check":QApplication.translate("uDisk","check",None)}

    _ItemPattern=re.compile("[0-9]?(.*)")
    
    def uniqueId(self):
        """
        renvoie un identifiant unique. Dans cette classe, cette fonction
        est synonyme de file()
        :return: un identifiant unique, garanti par le système de fichiers
        """
        return self.file()
        
    def headers(locale="C"):
        """
        Méthode statique, pour avoir des titres de colonne.
        renvoie des titres pour les items obtenus par __getitem__. 
        :param locale: la langue locale, pour traduire les titres éventuellement. Valeur par défaut : "C"
        :type  locale: str
        :return: une liste de titres de colonnes
        :rtype: list(str, ...)
        """
        result= list(uDisk2._specialItems.keys())+ list(uDisk2._itemNames.keys())
        return sorted(result)
        
    headers = staticmethod(headers)
    
    def __str__(self):
        """
        Fournit une représentation imprimable
        @return une représentation imprimable de l'instance
        """
        return self.title()+self.valuableProperties()

    def title(self):
        """
        Permet d'obtenir un identifiant unique de disque
        :return: le chemin dbus de l'instance
        """
        return self.path

    def isDosFat(self):
        """
        Permet de reconnaitre les partitions DOS-FAT
        :return: True dans le cas d'une partition FAT16 ou FAT32
        :rtype: bool
        """
        return self.fstype=="vfat"

    def isMounted(self):
        """
        :return: True si le disque ou la partion est montée
        :rtype: bool
        """
        return bool(self.mp)
        
    def valuableProperties(self,indent=4):
        """
        Facilite l'accès aux propriétés intéressantes d'une instance
        :return: une chaîne indentée avec les propriétés intéressantes, une par ligne
        :rtype: str
        """
        prefix="\n"+" "*indent
        r=""
        props=["mp", "parent", "fstype", "stickid", "uuid", "vendor", "model", "devStuff", "free", "capacity", "label"]
        for prop in props:
            r+=prefix+"%s = %s" %(prop, getattr(self,prop))
        return r

    def mountPoint(self):
        """
        :return: le point de montage
        :rtype: str
        """
        return self.mp

    def unNumberProp(self,n):
        """
        retire le numéro des en-têtes pour en faire un nom de propriété ;
        valide pour interroger dbus
        :param n: un numéro de propriété qui se réfère aux headers
        :type   n: int
        :return: une propriété renvoyée par dbus, dans un format imprimable
        :rtype: str
        """
        m=uDisk2._ItemPattern.match(self.headers()[n])
        try:
            return getattr(self, m.group(1))
        except:
            return ""
        
    def __getitem__(self,n):
        """
        Renvoie un élément de listage de données internes au disque
        :param n: un nombre
        :type  n: int
        :return: un élément si n>0, et le drapeau self.selected si n==0. 
        Les noms des éléments sont dans la liste itemNames utilisée dans 
        la fonction statique headers
        """
        propListe=self.headers()
        if n==0:
            return self.selected
        elif n <= len(propListe):
            return self.unNumberProp(n-1)

    def ensureMounted(self):
        """
        Permet de s'assurer qu'une partition ou un disque sera bien monté
        :result: le chemin du point de montage
        :rtype: str
        """
        mount_paths=self.mp
        if mount_paths==None: # le cas où la notion de montage est hors-sujet
            return ""
        leftTries=5
        while len(mount_paths)==0 and leftTries >0:
            leftTries = leftTries - 1
            path=self.path
            if len(path)>0:
                subprocess.call("udisks --mount %s > /dev/null" %path,shell=True)
                paths=self.mp
                print("STILL TO DEBUG: is the mount OK? is self.mp updated?")
                if paths:
                    return paths
                else:
                    time.sleep(0.5)
            else:
                time.sleep(0.5)
        if leftTries==0:
            raise Exception ("Could not mount the VFAT after 5 tries.")
        else:
            return mount_paths

            
        
class Available (UDisksBackend):
    """
    une classe pour représenter la collection des disques USB connectés

    les attributs publics sont :
    -  access le type d'accès qu'on veut pour les items
    -  targets la collection de disques USB, organisée en un dictionnaire de disques : les clés sont les disques, qui renvoient à un ensemble de partitions du disque
    -  firstFats une liste composée de la première partion DOS-FAT de chaque disque USB.

    """

    def __init__(self, access="disk", diskClass=uDisk2):
        """
        Le constructeur
        :param access: définit le type d'accès souhaité. Par défaut, c'est "disk"
          c'est à dire qu'on veut la liste des disques USB. Autres valeurs
          possibles : "firstFat" pour les premières partitions vfat.
        :param diskClass: la classe de disques à créer
        """
        UDisksBackend.__init__(self, diskClass=diskClass)
        self.access=access
        self.detect_devices()
        self.finishInit()

    def finishInit(self):
        """
        Fin de l'initialisation
        """
        self.mountFirstFats()

    def mountFirstFats(self):
        """
        fabrique la liste des partitions FAT,
        monte les partitions FAT si elles ne le sont pas
        """
        self.firstFats = self.getFirstFats()
        if self.access=="firstFat":
            for p in self.firstFats:
                uDisk2(p,self).ensureMounted()

    def __trunc__(self):
        """
        :return: le nombre de medias connectés ; permet d'appeler facilement int()
        """
        return len(self.firstFats)

    def compare(self, other):
        """
        Sert à comparer deux collections de disques, par exemple
        une collection passée et une collection présente.
        :param other: une instance de Available
        :return: vrai si other semble être la même collection de disques USB
        :rtype: bool
        """
        result=self.summary()==other.summary()
        return result

    def contains(self, ud):
        """
        Permet de déterminer si un disque est dans la collection
        :param ud: une instance de uDisk
        :return: vrai si le uDisk ud est dans la collection
        :rtype: bool
        """
        return ud.path in self.targets
    
    def disks(self):
        """
        Récolte les enregistrements de niveau supérieur de self.targets
        :return: la liste des chemins vers les disque USB détectés
        :rtype: list(str, ...)
        """
        return [d for d in self.targets if not self.targets[d].parent]

    def parts(self, d):
        """
        Récolte les partitions d'un disque
        :param d: le chemin vers un disque
        :return: la liste des partitions de ce disque
        :rtype: list(str, ...)
        """
        return [p for p in self.targets if self.targets[p].parent==d]

    def disks_ud(self):
        """
        Récolte les enregistrements de niveau supérieur de self.targets
        :return: la liste des objects uDisk2 détectés
        :rtype: list(uDisk2, ...)
        """
        return [self.targets[d] for d in self.targets if not self.targets[d].parent]

    def parts_ud(self, d):
        """
        Récolte les partitions d'un disque
        :param d: le chemin vers un disque
        :type  d: str
        :return: la liste des objets uDisk2 qui sont des partitions 
        de ce disque
        """
        return [self.targets[p] for p in self.targets if self.targets[p].parent==d]

    def parts_summary(self, d):
        """
        Fait un court résumé des partitions d'un disque
        :param d: le disque, donné soit sous forme udev soit comme chemin uDisks2
        :result: une liste ordonnée de descriptions courtes
        :rtype: list(str, ...)
        """
        def short(partition_ud):
            """
            Renvoie une description courte d'une partition ; abrège le label
            s'il est trop long.
           :@param partition_ud: une partition
            :return: une description
            :rtype: str
            """
            dev=os.path.basename(partition_ud.path)
            label=partition_ud.label
            label1=label.split(" ")[0]
            if len(label) > 11:
                label=label1[:8]+"..."
            if partition_ud.label:
                return "{}({})".format(dev, label)
            return dev
        d=os.path.join(
            "/org/freedesktop/UDisks2/block_devices", os.path.basename(d))
        return sorted([short(p) for p in self.parts_ud(d)])

    def summary(self):
        """
        Fournit une représentation imprimable d'un résumé
        :return: une représentation imprimable d'un résumé de la collection
        :rtype: str
        """
        r=  "Available USB disks\n"
        r+= "===================\n"
        for d in sorted(self.disks()):
            r+="%s\n" %(self.targets[d].devStuff)
            partlist=self.parts(d)
            if len(partlist)>0:
                r+="    Partitions :\n"
                for part in partlist:
                    r+="        %s\n" %(self.targets[part].devStuff,)
        return r

    def __str__(self):
        """
        Fournit une représentation imprimable
        :return: une représentation imprimable de la collection
        :rtype: str
        """
        r=  "Available USB disks\n"
        r+= "===================\n"
        for d in self.disks():
            r+="%s\n" %d
            partlist=self.parts(d)
            if len(partlist)>0:
                r+="    Partitions :\n"
                for part in sorted(partlist):
                    r+="        %s\n" %(self.targets[part].devStuff)
                    r+=self.targets[part].valuableProperties(12)+"\n"
        return r

    def disksDict(self):
        """
        :return: un dictionnaire disque => partitions
        """
        result={}
        for d in self.disks():
            result[d]=self.parts(d)
        return result

    def __getitem__(self, n):
        """
        Renvoye le nième disque. Le fonctionnement dépend du paramètre
        self.access
        :param n: un numéro
        :type  n: int
        :return: le nième disque USB connecté
        :rtype: uDisk2
        """
        if self.access=="disk":
            path=self.targets.keys()[n]
        elif self.access=="firstFat":
            path=self.firstFats[n]
        return self.targets[path]

    def __len__(self):
        """
        Renseigne sur la longueur de la collection. Le fonctionnement
        dépend du paramètre self.access
        :return: la longueur de la collection de disques renvoyée
        :rtype: int
        """
        if self.access=="disk":
            return len(self.targets)
        elif self.access=="firstFat":
            return len(self.firstFats)

    def getFirstFats(self):
        """
        Facilite l'accès aux partitions de type DOS-FAT, et a des effets de bord :
        - marque la première vfat dans chaque instance de disque
        - construit une liste des chemins uDisk des FATs

        :return: une liste de partitions, constituée de la première
        partition de type FAT de chaque disque USB connecté
        """
        result=[]
        disks=[d for d in self.targets if not self.targets[d].parent]
        for d in disks:
            parts=[p for p in self.targets if self.targets[p].parent==d]
            for p in parts:
                if self.targets[p].fstype=="vfat":
                    result.append(p)
                    # inscrit l'information dans l'instance du disque, par effet de bord
                    self.targets[d].firstFat=self.targets[p]
        return result

    def hasDev(self, dev):
        """
        :param dev: un chemin comme /org/freedesktop/UDisks/devices/sdb3
        :type  dev: str
        :return: True si la partition est dans la liste des partions disponibles
        :rtype: bool
        """
        s=str(dev)
        for p in self.fatPaths:
            if p.split("/")[-1]==s:
                return True
        return False
    
##################### fin de la définition de la calsse uDisk2 ################



#! /usr/bin/python3

import os, sys, argparse, glob, time, re
from subprocess import Popen, PIPE, call
from PyQt5.QtCore import QObject, QTranslator, QLocale
from PyQt5.QtWidgets import QApplication

class StickMaker(QObject):
    def __init__(self):
        QObject.__init__(self)
        self.startTime = time.time()
        self.pathSentence="PATH=/usr/sbin:/sbin:/usr/bin:/bin; "

    def hms(self):
        """
        returns the time elapsed so far, in format HH:MM:SS
        """
        t=int(time.time()-self.startTime)
        th=t//3600
        tmin=(t-3600*th)//60
        tsec=(t-3600*th-60*tmin)
        return f"{th:02d}:{tmin:02d}:{tsec:02d}"

    def flush(self, stderr=True, stdout=True):
        """
        flush standard outputs
        @param stderr don't flush sys.stderr if it's False (True by default)
        @param stdout don't flush sys.stdout if it's False (True by default)
        """
        if stderr: sys.stderr.flush()
        if stdout: sys.stdout.flush()
        return

    def log(self, s, timestamp=True):
        """
        print some information for stdout
        @param s a string to log
        @param timestamp True (by default) implies that there will be
        a timestamp
        """
        if timestamp:
            print(self.tr("[At {0}]    {1}").format(self.hms(),s))
        else:
            print(s)
        sys.stdout.flush()
        return

    def run(self):
        """
        The main method
        """
        if os.geteuid() != 0:
            self.log(
                self.tr("You must be root to launch {}").format(sys.argv[0]),
                timestamp=False
            )
            sys.exit(1)


        parser = argparse.ArgumentParser(description=self.tr('clone a Debian-Live distribution'))
        parser.add_argument('-f','--vfat-size', help=self.tr("size of the VFAT partition, in GB"))
        parser.add_argument('-n','--no-persistence-seed', help=self.tr("do'nt use the file rw.tar.gz to seed the persistence"), action='store_true')
        parser.add_argument('-s','--source', help=self.tr('source of the distribution (required); the keyword OWN means that we are cloning from a Debian-Live disk'), metavar="source", required=True)
        parser.add_argument('device', help=self.tr('the targetted USB stick'))
        args = parser.parse_args()

        ################### unmount targetted partitions if any #############

        cmd=f"mount | grep {args.device} | awk '{{print $1}}'"
        p=Popen(self.pathSentence+cmd, shell=True, stdout=PIPE, stderr=PIPE)
        mountpoints,_ = p.communicate()
        mounts=mountpoints.decode("utf-8").split()
        if mounts:
            self.log(self.tr("Unmounting {}").format(', '.join(mounts)))
        for m in mounts:
            call(self.pathSentence+f"umount {m}", shell=True)
        self.flush()

        ################### check the sizes #################################

        cmd = "lsblk| grep {}| grep disk| awk '{{print $4}}'".format(args.device.replace("/dev/",""))
        p = Popen(cmd, shell=True, stdout=PIPE)
        size, _ = p.communicate()
        size = size.decode("utf-8").strip().replace(",",".")
        assert(re.match(r"\d+(\.\d+)?G", size))
        size=float(size.replace("G",""))
        iso_size = os.stat(args.source).st_size//1024**3
        min_persistence = 1 # at least 1 GB for the persistence

        if iso_size + min_persistence > size:
            self.log(self.tr("The disk is too small ({} GiB) to contain the ISO image ({} GiB) plus at least 1 GB for the persistence").format(size, iso_size))
            assert (size > iso_size + min_persistence) # to stop the process

        if iso_size + float(args.vfat_size) + min_persistence > size:
            new_vfat_size = size - iso_size - min_persistence
            self.log(self.tr("The size of the VFAT partition has been shrinked down to {} GB").format(new_vfat_size))
            args.vfat_size = str(new_vfat_size)
        

        ################### copy the ISO-hybrid image #######################

        self.log(self.tr("Copying {0} to {1} ...").format(args.source, args.device))

        if args.source.upper()=="OWN":
            mountpoint = glob.glob("/usr/lib/live/mount/persistence/sd*1")[0]
            sourceDevice=mountpoint.replace("/usr/lib/live/mount/persistence","/dev")[:-1]
            ### when copying from own, the count of 4M blocks is computed
            ### from the last sector used by /dev/sd.1
            cmd=f"sfdisk -l {sourceDevice}| grep /dev/sd.1| awk '{{print $4}}'"
            p=Popen(self.pathSentence+cmd, shell=True, stdout=PIPE, stderr=PIPE)
            sectors,_ = p.communicate()
            sectors=int(sectors.decode("utf-8").strip())
            count = 1+sectors//8192 # convert 512 B sectors to 4 MiB blocks
            
            cmd = f"dd if={sourceDevice} of={args.device} status=progress bs=4M oflag=dsync count={count}"
        else:
            ### Then, the boot sector bears information for two partitions.
            cmd=f"dd if={args.source} of={args.device} status=progress bs=4M oflag=dsync"

        call(self.pathSentence+cmd, shell=True)
        self.flush()
        
        cmd=f"sync; sleep 4; partprobe {args.device}"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        inputs="" ### command lines for fdisk

        if args.source.upper()=="OWN":
            # if there are partitions #4 and #3 in the table, delete them
            for num in (4, 3):
                cmd=f"sfdisk -l {args.device}| grep {args.device}{num}"
                p=Popen(self.pathSentence+cmd, shell=True, stdout=PIPE, stderr=PIPE)
                reply, _ = p.communicate()
                if reply.decode("utf-8").strip(): # the partition #{num} does exist
                    self.log(self.tr("Delete the partition {}{}, to create it again").format(args.device, num))
                    inputs+=f"""\
d
{num}
"""

        ############# adds partition #3 for the VFAT data   ##########
        ############# adds partition #4 for the persistence ##########
        self.log(self.tr("Create the partition {}3 for VFAT data, using {}GiB").format(args.device, args.vfat_size))
        self.log(self.tr("Create the partition {}4 to support persistence").format(args.device))
        self.log(self.tr("This partition is adjusted to use all the remaining space on {}").format(args.device))
        inputs+="""\
n
p
3

+{}G
t
3
0b
n
p
4



w
""".format(args.vfat_size)
        p=Popen(self.pathSentence+f"fdisk {args.device}", shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        p.stdin.write(inputs.encode("utf-8"))

        message,err=p.communicate()
        self.log(message.decode("utf-8"))
        self.log(err.decode("utf-8"))

        cmd=f"sleep 6; partprobe {args.device}"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        ################ format VFAT   ####################################

        cmd = f"mkfs.vfat -n DATA {args.device}3"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        ################ format (ext4) ####################################

        cmd = f"mkfs.ext4 -L persistence -F {args.device}4"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        ################## mount the 4th partition ########################

        cmd = f"mkdir -p /tmp/{args.device}4; mount {args.device}4 /tmp/{args.device}4"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        ################# write the file persistence.conf ################

        self.log(self.tr("Create the file persistence.conf on the fourth partition"))
        with open(f"/tmp/{args.device}4/persistence.conf", "w") as outfile:
            outfile.write("/ union\n")

        ################## eventually seed the persistence partition ######

        if args.no_persistence_seed==False and os.path.exists("rw.tar.gz"):
            self.log(self.tr("Pre-seed the fourth partition with rw.tar.gz"))
            cmd = f"zcat rw.tar.gz | (cd /tmp/{args.device}4/; tar xf -)"
            call(self.pathSentence+cmd, shell=True)
        self.flush()

        #################### little cleanup ###############################

        cmd = f"umount /tmp/{args.device}4; rmdir /tmp/{args.device}4; partprobe {args.device}"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        cmd = f"sfdisk -l {args.device}"
        call(self.pathSentence+cmd, shell=True)
        self.flush()

        self.log(
            "\n=======================================================\n",
            timestamp=False
        )
        self.log(
            self.tr("Ready. Total time: {}").format(self.hms()),
            timestamp=False
        )

if __name__=="__main__":
    app = QApplication(sys.argv)
    # i18n stuff
    locale = QLocale.system().name()
    translation="live_clone_{}.ts".format(locale)
    langPath=os.path.join(os.path.abspath(os.path.dirname(__file__)),"lang",translation)
    translator = QTranslator(app)
    translator.load(langPath)
    app.installTranslator(translator)
    #############
    maker = StickMaker()
    maker.run()



"""
Defines wizards related to actions of live-clone
"""
from PyQt5.QtWidgets import QWizard, QWizardPage, QLineEdit, \
    QHBoxLayout, QVBoxLayout, \
    QRadioButton, QPlainTextEdit, QTreeView, QWidget, QCheckBox, \
    QMessageBox, QPushButton, QFileDialog, QLabel, QFrame
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import Qt

from tools import MountPoint, FileTreeSelectorModel
from monitor import Monitor, CloneMonitor

from subprocess import Popen, PIPE, call
from datetime import datetime
import os, re

def createVfatExt4Partitions(device, size="4G"):
    """
    create a VFAT and a EXT4 partition on a device.
    first find the number of the first VFAT partition, and erase any
    partition with this number or a greater number.
    Then create a VFAT partiton with size `size`, and a EXT4 partition
    to use all the space remaining on the device. Finally format both
    partitons according to their types
    @param device a string, like "/dev/sdc"
    @param size a string, defaults to "4G"
    @return the number of the last created partition (i.e. the EXT4 partition's
    number)
    """
    cmd="fdisk -l {}".format(device)
    p=Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
    out, err= p.communicate()
    out=out.decode("utf-8")
    FATlines=[]
    for l in out.split("\n"):
        m=re.match(r"^(/dev/\S+)\s.*FAT",l)
        if m:
            # remember the FAT partition and unmount it
            FATlines.append(m.group(1))
            call(f"umount {m.group(1)}", shell=True)
        else:
            m=re.match(r"^(/dev/\S+)\s.*",l)
            if m:
                # unmount other partitions too
                call(f"umount {m.group(1)}", shell=True)
    assert (FATlines) # there *must* be a first VFAT partition
    m=re.match(r"/dev/\D+(\d+)",FATlines[0])
    fatNum=int(m.group(1))
    inputs=f"""\
d

n
p
{fatNum}

+4G
t
0c
n
p
{fatNum+1}



w

"""
    p=Popen(f"fdisk {device}", shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)
    p.stdin.write(inputs.encode("utf-8"))
    message,err=p.communicate()
    call(f"sleep 6; partprobe {device}", shell=True)
    persistenceNum = fatNum+1
    call(f"mkfs.vfat -n DATA {device}{fatNum}", shell=True)
    call(f"mkfs.ext4 -L persistence -F {device}{persistenceNum}", shell=True)
    return persistenceNum

def new_packages(d, persistenceNum):
    """
    Work out a list of newer packages installed in the persistence
    zone of a debian-live Disk
    @param d a disk given as a /dev/sdx string
    @param persistenceNum the number of the persistence partition
    @return a text with on package by line
    """
    with MountPoint(d+"1") as base, MountPoint(d+str(persistenceNum)) as pers:
        cmd="""find "{pers}/rw/var/lib/dpkg/info" -name "*.list" -newer "{base}/live/filesystem.packages" | sed -e "s%{pers}/rw/var/lib/dpkg/info/%%" -e 's/.list$//' -e 's/:.*//'""".format(base=base, pers=pers)
        p=Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        out, err= p.communicate()
    return out.decode("utf-8")

class ToolWizard(QWizard):
    """
    The wizard associated to the "Tools" button of live-clone
    """

    def __init__(self, mw):
        """
        the constructor
        @param mw a MyMain (<= QMainWindow) instance
        """
        QWizard.__init__(self, mw)
        self.setWindowTitle(self.tr("Repair and/or manage a Freeduc USB stick"))
        page1=QWizardPage(self)
        page1.setTitle(self.tr("Select one disk"))
        page1.setSubTitle(self.tr("Which disk do you want to repair or to save?"))
        target=QLineEdit(page1)
        target.hide()
        page1.registerField("target*", target)
        layout=QVBoxLayout()
        radios={}
        def setTarget(checked):
            for r in radios:
                if r.isChecked():
                    page1.setField("target", r.text())
            return
        for d in mw.not_self_keys(format="udev"):
            rb=QRadioButton(d, parent=page1)
            rb.toggled.connect(setTarget)
            layout.addWidget(rb)
            radios[rb]=d
        page1.setLayout(layout)
        self.addPage(page1)
        ## assert page1.field("target")
        page2=QWizardPage(self)
        disk=self.field("target")
        persistenceNum=4 # default number for the persistence partition
        page2.setTitle(self.tr("List of additional packages in {}").format(disk))
        page2.setSubTitle(self.tr("Here is the list of packages which you can restore later. You can edit the text before saving it."))
        layout=QVBoxLayout()
        ed=QPlainTextEdit(page2)
        layout.addWidget(ed)
        newpackages=True
        def page2init(foo=None):
            nonlocal newpackages
            nonlocal persistenceNum
            disk=self.field("target")
            ## find the persistence partition if any
            diskname=disk.replace("/dev/","")
            cmd="ls -l /dev/disk/by-label| grep persistence | grep "+diskname
            p=Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
            out, err= p.communicate()
            if out:
                persistenceNum=out.decode("utf8").strip()[-1]
            else:
                newpackages=False
                return # no persistence, no need to develop the wizard further

            # we know the number of the persistence partition
            newpackages=new_packages(disk, persistenceNum)
            if not newpackages:
                packages=self.tr("There are no new packages installed in the persistence partition")
            packages=self.tr("""\
ADDITIONAL PACKAGES
===================
""") + "{}\n{}\n".format(disk,len(disk)*"-") + newpackages
            c=ed.textCursor()
            c.insertText(packages)
            ed.moveCursor(c.Start)
            ed.ensureCursorVisible()
            return
        page2.setLayout(layout)
        page2.initializePage=page2init
        def page2Validate(foo=None):
            if not newpackages:
                return True
            proposed_fname = "packages-{}.md".format(
                datetime.now().strftime("%Y-%m-%d--%H-%M")
            )
            fname, _ = QFileDialog.getSaveFileName(
                self, self.tr("Save Packages File"),
                proposed_fname,
                self.tr("Markdow files (*.md);;Plain text files (*.txt);;Any file (*)")
            )
            if fname:
                with open(fname,"w") as outfile:
                    outfile.write(ed.toPlainText())
                if "SUDO_UID" in os.environ and "SUDO_GID" in os.environ:
                    ## it is possible to downgrade the owner of the file
                    os.chown(
                        fname,
                        int(os.environ["SUDO_UID"]),
                        int(os.environ["SUDO_GID"])
                    )
            return True
        page2.validatePage=page2Validate
        self.addPage(page2)
        page3=QWizardPage(self)
        page3.setTitle(self.tr("Select personal data to save"))
        page3.setSubTitle(self.tr('Please check the tree of data which you want to backup. If you check the topmost node "user", all and every personal data is chosen, including Mozilla cache and desktop preferences. You may prefer to check subtrees. The subtrees are taken in account only when no node is checked above.'))
        layout=QVBoxLayout()
        tree=QTreeView(page3)
        layout.addWidget(tree)
        mp=None
        model=None
        exists=False
        def page3Init(foo=None):
            nonlocal mp, model, exists
            mp=MountPoint(self.field("target")+str(persistenceNum))
            mp.enter_()
            if mp:
                rootpath=os.path.join(str(mp), "rw", "home")
                exists = os.path.exists(rootpath)
                if exists:
                    model = FileTreeSelectorModel(rootpath=rootpath)
                    model.setRootPath(rootpath)
                    tree.setModel(model)
                    tree.setRootIndex(model.parent_index)

                    tree.setAnimated(False)
                    tree.setIndentation(20)
                    tree.setSortingEnabled(True)
                else:
                    ## there are no personal data
                    model = QStandardItemModel()
                    model.setHorizontalHeaderLabels(
                        [self.tr('Nothing is available')])
                    tree.setModel(model)
                    tree.setUniformRowHeights(True)

                    node1 = QStandardItem(
                        self.tr('No personal data in the persistence partition'))
                    model.appendRow(node1)
            return
        def page3validate(foo=None):
            nonlocal mp, exists
            if not exists:
                return True
            to_save=[]
            def cb(i, stack):
                if model.checkState(i)==Qt.Checked:
                    pathdata=[index.data() for index in stack+[i]]
                    to_save.append(os.path.join(*pathdata))
                return
            i=model.parent_index
            model.traverseDirectoryWhileUnchecked(i,callback=cb)
            ### ask for a tgz file name and make a tarball into it
            proposed_fname = "DATA-{}.tgz".format(
                datetime.now().strftime("%Y-%m-%d--%H-%M")
            )
            fname=""
            if to_save:
                fname, _ = QFileDialog.getSaveFileName(
                    self, self.tr("Save personal data archive"),
                    proposed_fname,
                    self.tr("TAR-GZIP archives (*.tgz,*.tar.gz);;Any file (*)")
                )
            if fname:
                ### launch the command asynchronously with a Monitor
                command = "/usr/bin/tar"
                args=[
                    "czvf",
                    fname
                ] + [os.path.join(str(mp), "rw", p) for p in to_save]
                ### add a new tab and a monitor
                n=mw.ui.tabWidget.count()
                title=self.tr("Save personal persistence data to {}").format(fname)
                newWidget=QWidget()
                mw.ui.tabWidget.addTab(
                    QWidget(),
                    self.tr("Save to {}").format(fname)
                )
                rindex=mw.rindexFromDev(self.field("target"))
                if rindex==None:
                    return True
                # rindex is defined, it bears the row number for the device
                def makeFinishCallback():
                    return lambda: mp.exit_()
                m=Monitor(
                    command, args, mw, n, rindex,
                    finishCallback=makeFinishCallback(),
                    environ=os.environ,
                    createfiles=[fname],
                    actionMessage=self.tr("save-data"),
                    buttonMessage=self.tr("Close the log of the archive"),
                )
                mw.monitors[m]=self.field("target")
                m.start()
            return True
        page3.initializePage=page3Init
        page3.validatePage=page3validate
        page3.setLayout(layout)
        self.addPage(page3)
        page4=QWizardPage(self)
        page4.setTitle(self.tr("Select personal data to save"))
        page4.setSubTitle(self.tr('Now, you have carefully backuped sensistive data, you may erase the persistence; however, you can still cancel it.'))
        def page4validate(foo=None):
            if self.field("ok1") and  self.field("ok2"):
                pers=self.field("target")+str(persistenceNum)
                call ("umount {}".format(pers), shell=True)
                def page4CallBackMaker():
                    def callback():
                        nonlocal persistenceNum
                        must_format = False # do we have to format the device?
                        # first check whether the persistence partition exists
                        with MountPoint(self.field("target")+str(persistenceNum)) as mp:
                            if not mp.capacity:
                                must_format = True
                        # format the device if necessary
                        if must_format:
                            persistenceNum = createVfatExt4Partitions(self.field("target"))
                        # then create the persistence.conf file
                        with MountPoint(self.field("target")+str(persistenceNum)) as mp:
                            confFile=os.path.join(str(mp),"persistence.conf")
                            call("echo '/ union' > {}".format(confFile), shell=True)
                        QMessageBox.information(
                            self,
                            self.tr("Persistence erased"),
                            self.tr("Persistence data have been erased from {}").format(self.field("target")+str(persistenceNum))
                        )
                        return
                    return callback
                command="mkfs"
                args= [
                    "-F",
                    "-L",
                    "persistence",
                    "-t",
                    "ext4",
                    pers
                ]
                n=mw.ui.tabWidget.count()
                title=self.tr("Format the partition {}").format(pers)
                newWidget=QWidget()
                mw.ui.tabWidget.addTab(
                    QWidget(),
                    self.tr("Format the partition {}").format(pers)
                )
                rindex=mw.rindexFromDev(self.field("target"))
                if rindex==None:
                    return True                
                m=Monitor(command, args, mw, n, rindex,
                          actionMessage="Formating",
                          finishCallback=page4CallBackMaker()
                )
                mw.monitors[m]=self.field("target")
                m.start()
            return True
        layout=QVBoxLayout()
        request1CB=QCheckBox(self.tr("Yes, I am sure that I want to erase the persistence."))
        request2CB=QCheckBox(self.tr("Yes again, I am completely sure!"))
        page4.registerField("ok1", request1CB)
        page4.registerField("ok2", request2CB)
        layout.addWidget(request1CB)
        layout.addWidget(request2CB)
        page4.setLayout(layout)
        page4.validatePage=page4validate
        self.addPage(page4)
        return
    
class CloneWizard(QWizard):
    """
    The wizard associated to the "Clone" button of live-clone
    """

    def __init__(self, mw):
        """
        the constructor
        @param mw a MyMain (<= QMainWindow) instance
        """
        QWizard.__init__(self, mw)
        self.setWindowTitle(self.tr("Clone a Freeduc system to USB drives"))
        page1=QWizardPage(self)
        page1.setTitle(self.tr("Select the target disks"))
        page1.setSubTitle(self.tr("Please check the disks where you want to clone the system. Warning: all data exiting on those disks will be erased, you can still Escape from this process."))
        selectedDisks=QLineEdit()
        selectedDisks.setReadOnly(True)
        selectedDisks.hide() # not visible, it bears juste a calculated field
        page1.registerField("selectedDisks*", selectedDisks)
        boxes={}
        def updateSelectedDisks(checkState=-1):
            """
            callback for changes in checkboxes
            """
            selected=[d for b,d in boxes.items() \
                      if b.checkState()==Qt.Checked]
            sel=",".join(selected)
            page1.setField("selectedDisks", sel)
            if checkState==-1:
                #called before the wizard is initialized
                selectedDisks.setText(sel)
            return
        vfatLineEdit = QLineEdit("0")
        page1.registerField("vfat", vfatLineEdit)
        def updateVFATsize(size = "0"):
            """
            callback for changes VFAT size lineEdit widget
            """
            value = ""
            for char in size:
                try:
                    v = float(value+char)
                    value = value + char
                except:
                    break
            size=value
            if size == "":
                size = "0"
            page1.setField("vfat", size)
            vfatLineEdit.setText(size)
            return
        vfatLineEdit.textEdited.connect(updateVFATsize)
        layout=QVBoxLayout()
        for disk in mw.not_self_keys(format="udev"):
            shortParts=mw.lesCles.parts_summary(disk)
            chb=QCheckBox("{} [{}]".format(
                disk,
                ", ".join(shortParts))
            )
            boxes[chb]=disk
            chb.stateChanged.connect(updateSelectedDisks)
            layout.addWidget(chb)
        layout.addWidget(selectedDisks)
        layout.addStretch()
        fr = QFrame()
        frLayout = QHBoxLayout()
        frLayout.addWidget(QLabel(self.tr("Size of a VFAT partition (GB):")))
        frLayout.addWidget(vfatLineEdit)
        frLayout.addStretch()
        fr.setLayout(frLayout)
        layout.addWidget(fr)
        page1.setLayout(layout)
        self.addPage(page1)
        sourceEdit=QLineEdit()
        sourceEdit.setReadOnly(True)
        page2=QWizardPage(self)
        page2.registerField("source*", sourceEdit)
        ownClone=os.path.exists("/usr/lib/live/mount/persistence")
        def page2init(foo=None):
            if ownClone:
                sourceEdit.setText("AUTOCLONE (Freeduc disks can clone themselves natively)")
            return
        page2.initializePage=page2init
        page2.setTitle(self.tr("Source of the core system"))
        if ownClone:
            page2.setSubTitle(self.tr("The program is running from a Freeduc GNU-Linux system"))
        else:
            page2.setSubTitle(self.tr("The program is running from a plain GNU-Linux system"))
        layout=QVBoxLayout()
        page2.setLayout(layout)
        def chooseISO(event=None):
            source, _ = QFileDialog.getOpenFileName(
                caption = self.tr("Choose an image to clone"),
                filter = self.tr("ISO Images (*.iso);;All files (*)")
            )
            page2.setField("source", source)
        chooseButton=QPushButton("Choose the source of the core system")
        chooseButton.clicked.connect(chooseISO)
        layout.addWidget(chooseButton)
        layout.addWidget(sourceEdit)
        def page2validate(foo=None):
            command="/usr/bin/python3"
            source=page2.field("source")
            if source.startswith("AUTOCLONE"):
                source="own"
            for device in self.field("selectedDisks").split(","):
                rindex=mw.rindexFromDev(device)
                if rindex==None:
                    return True
                args=[
                    "{}/makeLiveStick.py".format(mw.wd),
                    "--vfat-size",
                    vfatLineEdit.text(),
                    "--no-persistence-seed",
                    "--source",
                    source,
                    device
                ]
                ### add a new tab and a monitor
                n=mw.ui.tabWidget.count()
                title=self.tr("Cloning to {}").format(device)
                mw.ui.tabWidget.addTab(
                    QWidget(),
                    self.tr("Cloning to {}").format(device)
                )
                m=CloneMonitor(command, args, mw, n, rindex, environ=os.environ)
                mw.monitors[m]=device
                m.start()
            return True
        page2.validatePage=page2validate
        self.addPage(page2)
        return

class RunWizard(QWizard):
        """
        wizard for the runButton of live-clone
        """
        def __init__(self, mw, devices, available):
            """
            the constructor
            @param mw a MyMain (<= QMainWindow) instance
            @param devices a list of paths to USB devices
            @param available an Available instance
            """
            QWizard.__init__(self, mw)
            disks=sorted([d for d in devices])
            self.setWindowTitle(self.tr("Run a disk in a virtual machine"))
            page1=QWizardPage(self)
            page1.setTitle(self.tr("Select one disk"))
            page1.setSubTitle(self.tr("Which disk do you want to launch in a virtual machine?"))
            layout=QVBoxLayout()
            radios={}
            first=True
            for d in disks:
                rb=QRadioButton(d, parent=page1)
                if first:
                    rb.setChecked(True)
                    first=False
                layout.addWidget(rb)
                radios[rb]=d
            disks=QLineEdit(page1)
            disks.setReadOnly(True)
            disks.hide()
            def page1validate(foo=None):
                foundDisks=[]
                for rb in radios:
                    if rb.isChecked():
                        foundDisks.append(radios[rb])
                disks.setText(",".join(foundDisks))
                return bool(foundDisks)
            page1.validatePage=page1validate
            page1.registerField("disks", disks)
            page1.setLayout(layout)
            self.addPage(page1)
            return




import sys
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import datetime, os, re

class Monitor(QObject):
    def __init__(
            self,
            command, args,
            mainwindow, index, rindex,
            buttonMessage=None,
            actionMessage=None,
            createfiles=[],
            finishCallback=None,
            environ=None
    ):
        """
        the constructor
        @param command an executable
        @param args the command line args for the executable
        @param mainwindow a QWidget which should have signals monitorfinished
        and monitorclosed, and a QTabWidget in mainwindow.ui.tabWidget
        @param index the index of the wanted tab
        @param rindex the row index to access the table of devices
        @param buttonMessage the text for the button which appears at
        the end of the process
        @param actionMessage a short word to define the ongoing process
        @param createfiles a list of files which might be created: one can
        change their ownership if we are running as SUDO
        @param finishCallback a callback function which should be activated
        when the monitor finishes. This function takes no argument.
        @param environ a shell environment; it can be user to get variables
        SUDO_UID and SUDO_GID
        """
        QObject.__init__(self)
        self.category = "simple_monitor"
        self.index=index
        self.rindex=rindex
        self.command=command
        self.args=args
        self.emit_err = False
        self.emit_out = False
        if buttonMessage==None:
            buttonMessage=self.tr("Close this monitor")
        self.buttonMessage=buttonMessage
        if actionMessage==None:
            actionMessage=self.tr("Working")
        self.actionMessage=actionMessage
        self.createfiles=createfiles
        self.finishCallback=finishCallback
        self.environ=environ
        self.w=mainwindow.ui.tabWidget.widget(index)
        mainwindow.ui.tabWidget.setCurrentIndex(index)
        self.table=mainwindow.ui.tableWidget
        self.mainwindow=mainwindow
        self.process = QProcess(self.w)
        self.process.readyReadStandardOutput.connect(self.readOutput)
        self.process.readyReadStandardError.connect(self.readErr)
        self.process.finished.connect(self.finish)
        self.stdoutTerminal = QPlainTextEdit(self.w)
        self.stderrTerminal = QPlainTextEdit(self.w)
        self.cursor=QTextCursor(self.stdoutTerminal.document());
        self.err_cursor=QTextCursor(self.stderrTerminal.document());
        self.layout = QVBoxLayout(self.w)
        self.layout.addWidget(QLabel(self.tr("stdout")))
        self.layout.addWidget(self.stdoutTerminal, stretch=4)
        self.layout.addWidget(QLabel(self.tr("stderr")))
        self.layout.addWidget(self.stderrTerminal, stretch=1)
        return

    def start(self):
        self.process.start(self.command, self.args)
        # updates row rindex in self.table
        self.mainwindow.newmonitor.emit(self)
        return

    def downgrade_user(self):
        if self.environ and \
           "SUDO_UID" in self.environ and \
           "SUDO_GID" in self.environ:
            ## it is possible to downgrade the owner of the file
            for fname in self.createfiles:
                os.chown(
                    fname,
                    int(os.environ["SUDO_UID"]),
                    int(os.environ["SUDO_GID"])
                )
        self.createfiles=[]
        return

    def close(self):
        d=datetime.datetime.now()
        fname, _ =QFileDialog.getSaveFileName(
            self.mainwindow,
            self.tr("Save the messages?"),
            "{action}-log-{date}.log".format(
                action=self.actionMessage,
                date=d.strftime("%Y-%m-%d--%H-%M")
            ),
            self.tr("Log files (*.log);;All files (*)")
        )
        if fname:
            self.createfiles.append(fname)
            with open(fname, "w") as outfile:
                outfile.write("# STDOUT #\n")
                outfile.write(self.stdoutTerminal.toPlainText())
                outfile.write("\n# STDERR #\n")
                outfile.write(self.stderrTerminal.toPlainText())
        self.downgrade_user()
        self.mainwindow.monitorclosed.emit(self)
        return True
        
    def finish(self):
        button=QPushButton(self.buttonMessage)
        button.clicked.connect(self.close)
        self.layout.addWidget(button)
        self.downgrade_user()
        if self.finishCallback:
            self.finishCallback()
        self.mainwindow.monitorfinished.emit(self)

    def read(self, source, cursor, edit):
        """
        reads a QByteArray from source and writes it into a QPlainTextEdit
        widget
        @param source: either an instance of QProcess.readAllStandardOutput
        or an instance of QProcess.readAllStandardError
        @param cursor a QTextCursor instance
        @param edit a QPlainTextEdit instance
        @return the text line which was just read
        """
        text= bytes(source()).decode("utf-8")
        if text.startswith("\r"):
            ## carriage return
            cursor.select(cursor.LineUnderCursor)
            cursor.removeSelectedText()
            cursor.insertText(text[1:])
        else:
            cursor.insertText(text)
        edit.verticalScrollBar().setValue(edit.verticalScrollBar().maximum())
        return text
    
    def readOutput(self):
        text = self.read(self.process.readAllStandardOutput,
                         self.cursor,
                         self.stdoutTerminal)
        if self.emit_out:
            self.mainwindow.monitorout.emit(self,text)
        return
    
    def readErr(self):
        text = self.read(self.process.readAllStandardError,
                         self.err_cursor,
                         self.stderrTerminal)
        if self.emit_err:
            self.mainwindow.monitorerr.emit(self,text)
        return
    
# a pattern for the output of dd when copying the ISO image
# this pattern reports group(1) = bytes copied so far,
# group(2) = duration in seconds, group(3) = copy speed in GB/s
ddPattern=re.compile(r"^(\d+).*\(.*\).*, (\d+) s, ([.,\d]+) .*/s *$")

class CloneMonitor(Monitor):
    def __init__(self, command, args, mainwindow, index, rindex, environ=None):
        """
        the constructor
        @param command an executable
        @param args the command line args for the executable
        args[3] is the source to clone, args[4] is the targetted device
        @param mainwindow a QWidget which should have signals endmonitor
        and closemonitor, and a QTabWidget in mainwindow.ui.tabWidget
        @param index the index of the wanted tab
        @param rindex the row index to access the table of devices
        @param environ a shell environment; it can be user to get variables
        SUDO_UID and SUDO_GID
        """
        
        Monitor.__init__(self, command, args, mainwindow, index, rindex,
                         environ=environ
        )
        self.category = "clone_monitor"
        self.iso_filename = self.args[self.args.index("--source")+1]
        self.iso_filesize = os.stat(self.iso_filename).st_size
        self.buttonMessage=self.tr("Cloning is over")
        self.actionMessage=self.tr("Cloning")
        self.source=args[3]
        self.device=args[4]
        # this will trigger monitor_err events in the main window
        self.emit_err = True 
        return

    def progresstext(self, text):
        """
        given a text which comes from stderr, works out a progress message
        related to dd's work
        @param text a text issued by dd
        @return a progress message
        """
        def to_hms(t):
            t = int(t)
            h = t // 3600
            t = t - 3600 * h
            m = t // 60
            s = t - 60 * m
            return "{h:02d}:{m:02d}:{s:02d}".format(h=h, m=m, s=s)
        if "\r" in text:
            lines = [l for l in text.split("\r") if l]
            if lines:
                text=lines[0]
            else:
                return ""
        m = ddPattern.match(text)
        if m:
            bc = int(m.group(1)) # bytes copied
            eta = (self.iso_filesize - bc) / bc * float(m.group(2))
            return "{percent:4.2f}%, encore {eta}".format(
                percent = bc/self.iso_filesize*100,
                eta = to_hms(eta)
            )
        else:
            return ""

    
if __name__ == "__main__":
    app = QApplication(sys.argv)
    main = Monitor('ls', ['-lR', '../..'])
    main.show()
    sys.exit(app.exec_())



"""
Some routines to implement the Tools actions
"""
from subprocess import call, Popen, PIPE
from PyQt5.QtWidgets import QMessageBox, QFileSystemModel, QVBoxLayout, \
    QTreeView, QWidget, QPushButton, QDialog, QTableWidgetItem
from PyQt5.QtCore import QObject, QDir, Qt
import gi
gi.require_version('UDisks', '2.0')
from usbDisk2 import uDisk2
import tempfile, os, re
from ui_selectBackupData import Ui_BackupDialog

class FileTreeSelectorModel(QFileSystemModel):
    """
    This code is influenced by a reply to the question
    https://stackoverflow.com/questions/51338059/qfilesystemmodel-qtreeview-traverse-model-filesystem-tree-prior-to-view-e
    """
    def __init__(self, parent=None, rootpath='/', verbose=False):
        QFileSystemModel.__init__(self, None)
        self.root_path      = rootpath
        self.checks         = {}
        self.nodestack      = []
        self.parent_index   = self.setRootPath(self.root_path)
        self.root_index     = self.index(self.root_path)

        self.setFilter(QDir.AllEntries | QDir.Hidden | QDir.NoDot)
        self.directoryLoaded.connect(self._loaded)
        self.verbose=verbose

    def log(self, *messages):
        if self.verbose:
            print(*messages)
        return

    def _loaded(self, path):
        self.log('_loaded', path)

    def data(self, index, role=Qt.DisplayRole):
        if role != Qt.CheckStateRole:
            return QFileSystemModel.data(self, index, role)
        else:
            if index.column() == 0:
                return self.checkState(index)

    def flags(self, index):
            return QFileSystemModel.flags(self, index) | Qt.ItemIsUserCheckable

    def checkState(self, index):
        if index in self.checks:
            return self.checks[index]
        else:
            return Qt.Unchecked # not checked by default

    def setData(self, index, value, role):
        if (role == Qt.CheckStateRole and index.column() == 0):
            self.checks[index] = value
            self.log('setData(): {}'.format(value))
            return True
        return QFileSystemModel.setData(self, index, value, role)

    def traverseDirectory(
            self,
            parentindex,
            callback=None,
            stack=[]
    ):
        """
        recurse the directories and files visible in the view.
        @param parentindex the index of the dir or file being recursed
        @param callback a function which will be called with args
        parentindex and stack
        @param stack a list of previously recursed indices
        """
        self.log('traverseDirectory():')
        callback(parentindex, stack)
        if self.hasChildren(parentindex):
            self.log('|children|: {}'.format(self.rowCount(parentindex)))
            for childRow in range(self.rowCount(parentindex)):
                childIndex = parentindex.child(childRow, 0)
                self.log('child[{}]: recursing'.format(childRow))
                self.traverseDirectory(childIndex, callback=callback, stack=stack+[parentindex])
        else:
            self.log('no children')
        return

    def traverseDirectoryWhileUnchecked(
            self,
            parentindex,
            callback=None,
            stack=[]
    ):
        """
        recurse the directories and files visible in the view, and
        stops recursion wherever an item is checked.
        @param parentindex the index of the dir or file being recursed
        @param callback a function which will be called with args
        parentindex and stack
        @param stack a list of previously recursed indices
        """
        self.log('traverseDirectory():')
        callback(parentindex, stack)
        if self.checkState(parentindex)==Qt.Unchecked and \
           self.hasChildren(parentindex):
            self.log('|children|: {}'.format(self.rowCount(parentindex)))
            for childRow in range(self.rowCount(parentindex)):
                childIndex = parentindex.child(childRow, 0)
                self.log('child[{}]: recursing'.format(childRow))
                self.traverseDirectoryWhileUnchecked(childIndex, callback=callback, stack=stack+[parentindex])
        else:
            self.log('no children')
        return

    def printIndex(self, index):
        self.log('model printIndex(): {}'.format(self.filePath(index)))


class MountPoint:
    """
    Implements a mount point which will be associated with a file system
    and can be either permanent or temporary; it can be used in "with" 
    statements, ensuring that the file system is mounted in inside the block
    controlled by the statement, and unmounted when the block is finished
    if it was not mounted previously.
    """
    def __init__(self, p):
        """
        the constructor
        @param p a single partition of a disk, 
        given either as a string (/dev/something or
        /org/freedesktop/UDisks2/block_devices/something)
        or as a uDisk2 instance
        """
        orgPath="/org/freedesktop/UDisks2/block_devices"
        if type(p) == str:
            if p.startswith(orgPath):
                self._path=p
            elif p.startswith("/dev/"):
                self._path=os.path.join(orgPath, os.path.basename(p))
            else:
                raise Exception("Not a partition path: {}".format(p))
            devStuff=os.path.join("/dev", os.path.basename(self._path))
            self._partition=uDisk2(self.path, device=devStuff)
        elif isinstance(p, uDisk2):
            # reread the current properties, which might be volatile
            self._partition=uDisk2(p.path, device=p.devStuff)
            self._path=p.path
        else:
            raise Exception("Expected some other argument or type : {} type={}".format(p, type(p)))
        self._capacity = self._partition.capacity
        self._mountpoint=None
        self._temp=None
        return

    def __str__(self):
        return self._mountpoint

    def __repr__(self):
        return "MountPoint(path={_path}, dir={_mountpoint})".format(**self.__dict__)

    @property
    def path(self):
        return self._path

    @property
    def partition(self):
        return self._partition

    @property
    def mountpoint(self):
        return self._mountpoint

    @property
    def is_temporary(self):
        return self._temp

    @property
    def capacity(self):
        return self._capacity

    def enter_(self):
        pipe=Popen(
            "mount| grep {}".format(self._partition.devStuff),
            stdout=PIPE,
            stderr=PIPE,
            shell=True
        )
        out, err = pipe.communicate()
        out=out.decode("utf-8").strip()
        if out:
            m=re.match(r"{} on (.+) type .*".format(self._partition.devStuff), out)
            # the partition is mounted, no need to umount it
            self._mountpoint = m.group(1)
            self._temp = False
        elif self._partition.capacity: # only for existing partitions
            mp=""
            with tempfile.TemporaryDirectory(prefix="live_clone_") as temp:
                mp=temp
                # temp disappears when this block is over
            call("mkdir {}".format(mp), shell=True)
            call("mount {} {}".format(self._partition.devStuff, mp), shell=True)
            self._mountpoint = mp
            self._temp = True
        else:
            self._mountpoint = ""
        return
    
    def __enter__(self):
        """
        callback activated when one enters a With statement
        @return self
        """
        self.enter_()
        return self

    def exit_(self):
        if self._temp:
            # temporary mount points are dismissed
            call("umount {}".format(self._partition.devStuff), shell=True)
            call("rmdir {}".format(self._mountpoint), shell=True)
        return
    
    def __exit__(self, type, value, tb):
        """
        Callback to exit a With statement; unmounts temporary mounts
        @param type type of Exception if any
        @param value parameter of the exception
        @param tb an TraceBack instance
        @return True is no exception occured
        """
        self.exit_()
        return not tb    
        
                
class Tool(QObject):
    """
    This class is a subclass of QObject to inherit translation functions.
    real tools at its methods
    """
    
    def runDisk(self, d, window, memory="1G"):
        """
        run a disk in a virtual machine
        @param d a disk given as a uDisk2 instance
        @param window a QMainWindow instance
        @param memory the amount of RAM left to the virtual machine (default=1G)
        """
        cmd="(kvm -m {} -drive file={},format=raw &)".format(memory, d)
        call(cmd, shell=True)
        return

    def select_backup(self, d, window):
        """
        Work out a list of paths to directories one wants to backu
        @param d a disk given as a /dev/sdx string
        @param window a QMainWindow instance
         """
        ## the path to personal data is:
        ## <mount point>/rw/home
        ## the partition to mount is the partition #3
        with MountPoint(d+"3") as mp:
            rootpath=os.path.join(str(mp), "rw", "home")
            d=QDialog(window)
            d.ui=Ui_BackupDialog()
            d.ui.setupUi(d)
            d.model = FileTreeSelectorModel(rootpath=rootpath)
            d.model.setRootPath(rootpath)
            d.ui.tree.setModel(d.model)
            d.ui.tree.setRootIndex(d.model.parent_index)

            d.ui.tree.setAnimated(False)
            d.ui.tree.setIndentation(20)
            d.ui.tree.setSortingEnabled(True)

            d.ui.tree.setWindowTitle("Dir View")

            ok=d.exec_()
            to_save=[]
            def recurseFiles():
                def cb(i, stack):
                    if d.model.checkState(i)==Qt.Checked:
                        pathdata=[index.data() for index in stack+[i]]
                        to_save.append(os.path.join(*pathdata))
                    return
                i=d.model.parent_index
                d.model.traverseDirectoryWhileUnchecked(i,callback=cb)
                return
            if ok:
                recurseFiles()
            ### GRRR the result is in to_save
        return

    def erase_persistence(self, p, window):
        """
        Erase persistence data from a partition
        @param p a partition given as a uDisk2 instance
        @parm window a QMainWindow instance
        """
        if os.getuid()!= 0:
            # we are not root
            QMessageBox.critical(
                window,
                self.tr("Cannot erase persistence data"),
                self.tr("One needs root priviledge to erase persistence data")
            )
            return
        with MountPoint(p) as mp:
            # erase directories work and rw from the persistence partition
            call("cd {}; rm -rf work rw".format(mp), shell=True)
            confFile=os.path.join(str(mp),"persistence.conf")
            if not os.path.exists(confFile):
                # the file persistence.conf must be eventually restored
                call("cd {}; echo '/ union' > persistence.conf".format(mp), shell=True)
        QMessageBox.information(
            window,
            self.tr("Persistence erased"),
            self.tr("Persistence data have been erased from {}").format(p.devStuff)
        )
        return

class MyItem(QTableWidgetItem):
    """
    non-editable QTableWidgetItem items
    """
    def __init__(self, text):
        QTableWidgetItem.__init__(self,text)
        self.setFlags(self.flags() & ~Qt.ItemIsEditable)
        return

